/* SetFID.c -- Copyright 1989, 1993 Liam R. E. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

/* SetFID.c -- get/set the largest current word number
 * 
 * $Id: SetFID.c,v 1.2 93/11/28 15:15:26 lee Exp $
 */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>

#include "fileinfo.h"
#include "wordinfo.h"
#include "blkheader.h"
#include "block0.h"

/** declarations: **/
/** Unix system calls that need to be declared: **/

/** Unix Library Calls that need to be declared: **/

/** lq-text calls that need to be declared: **/
extern unsigned char *ReadBlock();
extern void WriteBlock(), FlushCache();
extern t_FID GetMaxFID();

#ifdef ASCIITRACE
 extern int AsciiTrace;
#endif
/** **/

extern t_FID LastNextFIDVal;

#undef GetNextFID

t_FID GetNextFID();

LIBRARY t_FID
SpoofGetNextFID(WriteCurrent)
{
    static int SinceLastUpdate = 0;

    if (WriteCurrent) {
	SinceLastUpdate = 0;
	return GetNextFID(WriteCurrent);
    }

    /* Call the real function sometimes, so that the database does
     * get updated in case of a crash or for other users.
     */
    if (++SinceLastUpdate > 10000) {
	SinceLastUpdate = 0;
	(void) GetNextFID(1);
	/* GetNextFID(1) sets LastNextFIDVal but doesn't increment it;
	 * This is ugly and has to be made cleaner... TODO FIXME
	 */
    }

    if (LastNextFIDVal == (t_FID) 0) {
	LastNextFIDVal = GetMaxFID();
	SinceLastUpdate = 0;
    }
    ++LastNextFIDVal;
    return LastNextFIDVal;
}

API void
WriteCurrentMaxFID()
{
    (void) GetNextFID(1);
}

/*ARGSUSED1*/
API t_FID
GetNextFID(FileSize)
    long FileSize;
{
    t_FID Result = 0;
    unsigned char *p;
    t_BlockHeader *BH;

    /** Alter the file, so other programs can see the new words...
     **/

    p = ReadBlock(0L);

    Result = (t_FID) p[BLOCK_ZERO_MAXFID0]; /* most significant byte */
    Result <<= 8; Result |= (t_FID) p[BLOCK_ZERO_MAXFID1];
    Result <<=8; Result |= (t_FID) p[BLOCK_ZERO_MAXFID2];
    Result <<=8; Result |= (t_FID) p[BLOCK_ZERO_MAXFID3];

    /* if WriteCurrent is set, we should not increment anything */
    ++LastNextFIDVal;
    ++Result;

    if (LastNextFIDVal > Result) {
	Result = LastNextFIDVal;
    } else {
	LastNextFIDVal = Result;
    }

    {
	t_FID tmp = Result;

	p[BLOCK_ZERO_MAXFID3] = (tmp & 0xff); /* least significant byte */
	tmp >>= 8; p[BLOCK_ZERO_MAXFID2] = (tmp & 0xff);
	tmp >>= 8; p[BLOCK_ZERO_MAXFID1] = (tmp & 0xff);
	tmp >>= 8; p[BLOCK_ZERO_MAXFID0] = (tmp & 0xff);
    }

    BH = (t_BlockHeader *) p;
    BH->NumberOfBlocks = 1;
    WriteBlock(0L, p, 1); /* leave it in the cache */

#ifdef ASCIITRACE
    if (AsciiTrace > 6) {
	fprintf(stderr, "Largest File Identifier (\"FID\") now %ld\n", Result);
    }
#endif

    return Result;
}

