/* gettitle.c -- Copyright 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file COPYRIGHT for full details.
 */

#ifndef LINE
static char *RcsId = "@(#) $Id: gettitle.c,v 1.2 93/12/09 21:51:30 lee Exp $";
#endif

/* This file defines:
 *
 * char *
 * GetDocumentTitle(t_FID FID)
 *
 */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include <stdio.h> /* stderr, also for fileinfo.h */
#include <fcntl.h>
#include <malloc.h>
#include <sys/types.h>
#include "fileinfo.h" /* for t_FID.h */

#include "emalloc.h"

#ifndef SEEK_SET
# include <unistd.h> /* for SEEK_SET, or use 0 if it's not found */
#endif

/** Unix system calls that need to be declared: **/

/** C library functions that need to be declared: **/
extern char *memcpy();

/** lqtext library functions that need to be declared: **/

/** Functions within this file that need to be declared: **/

/** **/

#ifdef ASCIITRACE
extern int AsciiTrace;
#endif

static char *titleFileName = 0;
static int titleFile;
typedef enum {
    FS_NotTried, FS_Open, FS_Closed, FS_NotAvailable
} t_FileState;

static t_FileState FileState = FS_NotTried;

static int
OpenTitleFile()
{
    extern char *joinstr3();

    titleFileName = joinstr3(DatabaseDir, "/", "titles");

    titleFile = open(titleFileName, O_RDONLY, 0);

    if (titleFile < 0) {
	Error(E_WARN|E_SYS,
	    "couldn't open document title file %s", titleFileName
	);
    }
    return titleFile;
}

static char *ReadTitle();

API char *
GetDocumentTitle(FID, Name)
    t_FID FID;
    char *Name;
{
    switch (FileState) {
    case FS_NotAvailable:
	return Name;
    case FS_Closed:
    case FS_NotTried:
	if (OpenTitleFile() < 0) {
	    FileState = FS_NotAvailable;
	    return Name;
	} else {
	    FileState = FS_Open;
	}
	/* fall through */
    case FS_Open:
    default:
	/* fall through */
	break;
    }

    /* the file is open, let's read the title */
    return ReadTitle(FID, Name);

}

#define MAXLINE 4096
static char *
ReadTitle(FID, Name)
    t_FID FID;
    char *Name;
{
    static char Buffer[MAXLINE];
    long LowestPossibleOffsetOfStartOfLine = 0L;
    long HighestPossibleOffsetOfStartOfLine;
    int MinimumGapBetweenLines;
    struct stat statbuf;
    long ValueFound = 0;

    /* the format is
     * FID	title
     * where FID is an ascii hexadeciml representation,
     * and title is a string.
     * TODO: use data file to store this.
     */
    
    if (fstat(titleFile, &statbuf) < 0) {
	Error(E_WARN|E_SYS,
	    "couldn't open file \"%s\" containing Document Titles (fstat %d)",
	    titleFileName,
	    titleFile
	);
	FileState = FS_NotAvailable;
	if (titleFile >= 0) {
	    (void) close(titleFile);
	}
	return Name;
    }
    
    /*** binary search ***/

    /* work out how close the lines might be.  Since the line we
     * want must contain the FID and a newline, we'll use strlen
     * as a guess:
     */
    (void) sprintf(Buffer, "%ld", FID);
    MinimumGapBetweenLines = strlen(Buffer) + 1;

    /* the line can't start any nearer the end of the file than
     * MinimumGapBetweenLines:
     */
    HighestPossibleOffsetOfStartOfLine = statbuf.st_size -
							MinimumGapBetweenLines;
    
    for (;;) {
	long Guess;
	int BytesRead;
	register char *p;

NextGuess:
	if (LowestPossibleOffsetOfStartOfLine >
				HighestPossibleOffsetOfStartOfLine) {
	    break;
	}

	Guess = (HighestPossibleOffsetOfStartOfLine +
		LowestPossibleOffsetOfStartOfLine) / 2;
	if (Guess < MinimumGapBetweenLines) {
	    Guess = 0;
	}

SeekAgain:

	if (lseek(titleFile, Guess, SEEK_SET) < 0) {
	    Error(E_WARN|E_SYS,
		"couldn't lseek to byte %ld in document title file \"%s\"",
		Guess,
		titleFileName
	    );
	    FileState = FS_Closed; /* but try again */
	    (void) close(titleFile);
	    return Name;
	}

ReadMore:
	BytesRead = read(titleFile, Buffer, MAXLINE);

	if (BytesRead <= 0) {
	    Error(E_SYS|E_WARN,
		"couldn't read from document title file \"%s\" at offset %ld",
		titleFileName,
		Guess
	    );
	    FileState = FS_Closed; /* but try again */
	    (void) close(titleFile);
	    return Name;
	}

	ValueFound = 0;
	for (p = Buffer; p - Buffer < BytesRead - MinimumGapBetweenLines; p++) {
	    if (*p == '\n' || (Guess == 0L && p == Buffer)) {

		/* got the start of a line */
		if (Guess != 0L) {
		    p++;
		}

		if (*p == '#') { /* comment ignored */
		    continue;
		} else if (!isdigit(*p)) {
		    Error(E_WARN,
"Document Title File \"%s\" has line not starting with digit or #",
			titleFileName
		    );
		    FileState = FS_NotAvailable;
		    (void) close(titleFile);
		    return Name;
		}

		ValueFound = 0;
		while (isdigit(*p) && p - Buffer < MAXLINE) {
		    ValueFound *= 10;
		    ValueFound += *p - '0';
		    p++;
		}

		if (p - Buffer >= MAXLINE - 2) { /* we need \t and \n */
		    Guess += MAXLINE / 2;
		    goto SeekAgain;
		}

		if (ValueFound < FID) {
		    continue;
		} else if (ValueFound > FID) {
		    HighestPossibleOffsetOfStartOfLine = Guess;
		    goto NextGuess;
		} else if (ValueFound == FID) {
		    char *Result = ++p;

		    while (p - Buffer <= MAXLINE) {
			if (*p == '\n') {
			    break;
			}
			p++;
		    }
		    *p = '\0';
		    return Result;
		}
	    }
	}
	if (ValueFound && ValueFound < FID) {
	    LowestPossibleOffsetOfStartOfLine = Guess + (p - Buffer);
	    if (ValueFound + 3 >= FID) {
		goto NextGuess;
	    }
	} else if (p - Buffer >= MAXLINE) {
	    Guess += MAXLINE / 2;
	    goto SeekAgain;
	}
    }
    return Name; /* not found */
}
